/* client.c */
/*
Copyright 2003 Aris Adamantiadis

This file is part of the SSH Library

You are free to copy this file, modify it in any way, consider it being public domain. This does not apply to the rest
of the library though, but is allowed so you can cut-and-paste working code from this file.
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>

#include <sys/select.h>
#include <sys/time.h>
#include <signal.h>
#include <errno.h>
#include <libssh/libssh.h>
#include <libssh/sftp.h>

#include <fcntl.h>

#define MAXCMD 10
char *host;
char *user;
int sftp;
char *cmds[MAXCMD];
struct termios terminal;


void add_cmd(char *cmd){
    int n;
    for(n=0;cmds[n] && (n<MAXCMD);n++);
    if(n==MAXCMD)
        return;
    cmds[n]=strdup(cmd);
}

void usage(){
    fprintf(stderr,"Usage : ssh [options] [login@]hostname\n"
    "Options :\n"
    "  -l user : log in as user\n"
    "  -p port : connect to port\n"
    "  -d : use DSS to verify host public key\n"
    "  -r : use RSA to verify host public key\n");
    exit(0);
}

int opts(int argc, char **argv){
    int i;
    if(strstr(argv[0],"sftp"))
        sftp=1;
//    for(i=0;i<argc;i++)
//        printf("%d : %s\n",i,argv[i]);
    /* insert your own arguments here */
    while((i=getopt(argc,argv,""))!=-1){
        switch(i){
            default:
                fprintf(stderr,"unknown option %c\n",optopt);
                usage();
        }
    }
    if(optind < argc)
        host=argv[optind++];
    while(optind < argc)
        add_cmd(argv[optind++]);
    if(host==NULL)
        usage();
    return 0;
}

void do_cleanup(){
    tcsetattr(0,TCSANOW,&terminal);
}

void select_loop(SSH_SESSION *session,CHANNEL *channel){
    fd_set fds;
    struct timeval timeout;
    char buffer[10];
    BUFFER *readbuf=buffer_new();
    CHANNEL *channels[]={channel,NULL};
    CHANNEL *outchannel[2];
    int lus;
    int eof=0;
    int interactive=isatty(0);
    struct termios terminal_local;
    
    if(interactive){
        tcgetattr(0,&terminal_local);
        memcpy(&terminal,&terminal_local,sizeof(struct termios));
        cfmakeraw(&terminal_local);
        tcsetattr(0,TCSANOW,&terminal_local);
    }
    signal(SIGTERM,do_cleanup);
    while(channel){
        FD_ZERO(&fds);
        if(!eof)
            FD_SET(0,&fds);
        timeout.tv_sec=30;
        timeout.tv_usec=0;
        ssh_select(channels,outchannel,0+1,&fds,&timeout);
        if(FD_ISSET(0,&fds)){
            lus=read(0,buffer,10);
            if(lus)
                channel_write(channel,buffer,lus);
            else{
                eof=1;
                channel_send_eof(channel);
            }
        }
        if(outchannel[0]){
            while(channel_poll(outchannel[0],0)){
                lus=channel_read(outchannel[0],readbuf,0,0);
                if(lus==-1){
                    ssh_say(0,"error reading channel : %s\n",ssh_get_error(session));
                    return;
                }
                if(lus==0){
                    ssh_say(1,"EOF received\n");
                } else
                    write(1,buffer_get(readbuf),lus);
            }
            while(channel_poll(outchannel[0],1)){ /* stderr */
                lus=channel_read(outchannel[0],readbuf,0,1);
                if(lus==-1){
                    ssh_say(0,"error reading channel : %s\n",ssh_get_error(session));
                    return;
                }
                if(lus==0){
                    ssh_say(1,"EOF received\n");
                } else
                    write(2,buffer_get(readbuf),lus);
            }
        }
        if(channel->open==0){
            channel_free(channel);
            channel=NULL;
        }
    }
    buffer_free(readbuf);
}

void shell(SSH_SESSION *session){
    CHANNEL *channel;
    channel = open_session_channel(session,1000,1000);
    if(isatty(0))
        channel_request_pty(channel);
    channel_request_shell(channel);
    select_loop(session,channel);
}

/* it's just a proof of concept code for sftp, till i write a real documentation about it */
void do_sftp(SSH_SESSION *session){
    SFTP_SESSION *sftp=sftp_new(session);
    SFTP_DIR *dir;
    SFTP_ATTRIBUTES *file;
    SFTP_FILE *fichier;
    SFTP_FILE *to;
    int len=1;
    char data[128];
    if(!sftp){
        ssh_say(0,"sftp error initialising channel : %s\n",ssh_get_error(session));
        return;
    }
    if(sftp_init(sftp)){
        ssh_say(0,"error initialising sftp : %s\n",ssh_get_error(session));
        return;
    }
    /* the connection is made */
    /* opening a directory */
    dir=sftp_opendir(sftp,"./");
    if(!dir) {
        ssh_say(0,"Directory not opened(%s)\n",ssh_get_error(session));
        return ;
    }
    /* reading the whole directory, file by file */
    while((file=sftp_readdir(sftp,dir))){
        ssh_say(0,"%30s(%.8lo) : %.5d.%.5d : %.10lld bytes\n",file->name,file->permissions,file->uid,file->gid,file->size);
        sftp_attributes_free(file);
    }
    /* when file=NULL, an error has occured OR the directory listing is end of file */
    if(!sftp_dir_eof(dir)){
        ssh_say(0,"error : %s\n",ssh_get_error(session));
        return;
    }
    if(sftp_dir_close(dir)){
        ssh_say(0,"Error : %s\n",ssh_get_error(session));
        return;
    }
    /* this will open a file and copy it into your /home directory */
    /* the small buffer size was intended to stress the library. of course, you can use a buffer till 20kbytes without problem */

    fichier=sftp_open(sftp,"/usr/bin/ssh",O_RDONLY,NULL);
    if(!fichier){
        ssh_say(0,"Error opening /usr/bin/ssh : %s\n",ssh_get_error(session));
        return;
    }
    /* open a file for writing... */
    to=sftp_open(sftp,"ssh-copy",O_WRONLY | O_CREAT,NULL);
    if(!to){
        ssh_say(0,"Error opening ssh-copy for writing : %s\n",ssh_get_error(session));
        return;
    }
    while((len=sftp_read(fichier,data,128)) > 0){
        if(sftp_write(to,data,len)!=len){
            ssh_say(0,"error writing %d bytes : %s\n",len,ssh_get_error(session));
            return;
        }
    }
    if(len<0)
        ssh_say(0,"Error reading file : %s\n",ssh_get_error(session));     
    sftp_file_close(fichier);
    sftp_file_close(to);
    /* close the sftp session */
    sftp_free(sftp);
}

int main(int argc, char **argv){
    SSH_SESSION *session;
    SSH_OPTIONS *options;
    int auth=0;
    char *password;
    char *banner;
    char hash[MD5_DIGEST_LEN];

    options=ssh_getopt(&argc, argv);
    if(!options){
        fprintf(stderr,"Error : %s\n",ssh_get_error(NULL));
        usage();
        return -1;
    }    
    opts(argc,argv);
    if(user)
        options_set_username(options,user);
    options_set_host(options,host);
    session=ssh_connect(options);
    if(!session){
        fprintf(stderr,"Connection failed : %s\n",ssh_get_error(NULL));
        return -1;
    }

    /* don't you remember you hash value by heart ? */
    pubkey_get_hash(session,hash);
    ssh_print_hexa("Public key hash",hash,MD5_DIGEST_LEN);
    /* no ? you should :) */
    
    auth=ssh_userauth_autopubkey(session);
    if(auth==AUTH_ERROR){
        fprintf(stderr,"Authenticating with pubkey: %s\n",ssh_get_error(session));
        return -1;
    }
    banner=ssh_get_issue_banner(session);
    if(banner){
        printf("%s\n",banner);
        free(banner);
    }
    if(auth!=AUTH_SUCCESS){
        password=getpass("Password : ");
        if(ssh_userauth_password(session,NULL,password) != AUTH_SUCCESS){
            fprintf(stderr,"Authentication failed: %s\n",ssh_get_error(session));
            ssh_disconnect(session);
            return -1;
        }
        memset(password,0,strlen(password));
    }
    ssh_say(1,"Authentication success\n");
    if(!sftp)
        shell(session);
    else
        do_sftp(session);    
    if(!sftp)
        do_cleanup();
    ssh_disconnect(session);
    return 0;
}
